/****************************************************************************
 *
 *   File: timers.cxx:
 *   
 *   Author: Mark Stephenson
 *   Date: Thu Oct  9 18:09:52 2003
 *
 *   Function: This library contains global variables that the loop
 *      instrumentor uses to instrument loops.  The functions 
 *      __measure_loop_init and __measure_loop_final initialize and
 *      finalize the library.  Other than that, the loop instrumentor
 *      inserts all the requisite code.  It should be noted however,
 *      that the inserted code accesses the global variables defined
 *      in this object.
 *
 ***************************************************************************/

#include "timers.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define NO_PROBLEMS -1

// The loop timer array contains 2*N entries, where N is the number
// of loops that are instrumented.  If I is the id of a particular
// loop, the loop's running counter is the 2*Ith index, and the
// cumulative total is in 2*I+1.
unsigned long int *_loop_timers;

// This variable keeps track of the number of timers (i.e., the
// number of instrumented loops.
int _loop_timers_size = 0;

// This variable is used as an error condition code register.
// If its value when __measure_loop_final is greater than or
// equal to NO_PROBLEMS, then there was a problem with the 
// associated loop.
int _loop_timers_error = NO_PROBLEMS;

// This variable holds the name of the benchmarks you're running.
char *_loop_exec_name;

#define assert_die(cond) \
if (!(cond)) { fprintf (stderr, "Assertion failure\n"); exit (-1); }

// ------------------------------------------------------------------    
// Initializes the library.
// ------------------------------------------------------------------
void __measure_loop_init (char *fun_name, signed int id)
{
  int i;
  char gidfname[80];
  FILE *timers_size_file;

  _loop_exec_name = fun_name;
  _loop_timers_error = NO_PROBLEMS;

  fprintf (stderr, 
	   "LOOP_INSTR_LIB: Calling __measure_loop_init from %s...\n",
	   fun_name);

  timers_size_file = fopen ("last_instr_gid", "r");
  if (!timers_size_file)
  {
    sprintf (gidfname, "gid.%s", fun_name);
    timers_size_file = fopen (gidfname, "r");
  }

  assert_die (timers_size_file);
  fscanf (timers_size_file, "%d", &_loop_timers_size);
  assert_die (_loop_timers_size >= 0);

  fprintf (stderr, "LOOP_INSTR_LIB: Allocating %d timers...\n",
	   _loop_timers_size);

  _loop_timers = (unsigned long int*) 
    malloc (_loop_timers_size * 2 * sizeof(unsigned long int));
  assert_die (_loop_timers);
  
  for (i = 0; i < _loop_timers_size*2; i++)
  {
    _loop_timers[i] = 0;
  }
}


// ------------------------------------------------------------------    
// Finalizes the library.
// ------------------------------------------------------------------
void __measure_loop_final (char *fun_name, signed int id)
{
  int i;
  char unique_fname[80];
  FILE *unique_file;

  strcpy(unique_fname, _loop_exec_name);
  strcat(unique_fname, ".XXXXXX");
  
  assert_die (mktemp(unique_fname));

  if (_loop_timers_error != NO_PROBLEMS)
  {
    fprintf (stderr, "LOOP_INSTR_LIB: bad condition code (%d).\n",
	     _loop_timers_error);
    exit (-1);
  }

  unique_file = fopen (unique_fname, "w");
  assert_die (unique_file);

  fprintf (stderr, 
	   "LOOP_INSTR_LIB: Calling __measure_loop_final from %s...\n",
	   fun_name);
  for (i = 0; i < _loop_timers_size; i++)
  {
    if (_loop_timers[2*i] != 0) 
    {
      fprintf (stderr, 
	       "LOOP_INSTR_LIB, Error: start without stop (%d).\n", i);
      exit (-1);
    }

    fprintf (unique_file, "%d %ld\n", i, _loop_timers[2*i+1]);
  }

  fclose (unique_file);
  _loop_timers_size = 0;
  free (_loop_timers);
}



